/************************************************************************************
 * Produto Final
 ************************************************************************************/
/*
 * Includes
 */
#include <stdio.h>		// printf(), file, ...
#include <sys/ipc.h>            // Todas as funcoes de IPC
#include <sys/msg.h>		// msgget(), msgsnd(), msgrcv()
#include <sys/shm.h>            // shmget(), shmat()
#include <sys/sem.h>            // semget(), semop()
#include <pthread.h>		// pthread()

/*
 * Constantes
 */
#define MSQ_CID		0x1111	// Chave da Fila de Mensagens com a Camada Independete de Dispositivo
#define MSQ_CTI		0x1222	// Chave da Fila de Mensagens com a Camada Tratamento de Interrupcoes
#define SHM_REGCTRL	0x2111	// Chave da Memoria Compartilhada dos Registradores da Controladora
#define SHM_TABCFG	0x2222	// Chave da Memoria Compartilhada da Tabela de Configuracao
#define SHM_MPART	0x2333	// Chave da Memoria Compartilhada da Matriz de Particoes
#define SEM_PROD	0x3111	// Chave do Semaforo de Exclusao Mutua entre Produtores
#define SEM_LISTA	0x3222	// Chave do Semaforo de Exclusao Mutua da Lista de Requisicoes
#define SEM_CONTLISTA	0x3333	// Chave do Semaforo Contador de Requisicoes da Lista de Requisicoes



/*
 * Variaveis Globais
 */
struct sembuf	down[1] = {0,-1,0};	// Estrutura para operacao de decremento do semaforo
struct sembuf	up[1] = {0,1,0};	// Estrutura para operacao de incremento do semaforo

int msqid_cid;				// ID da Fila de Mensagens com a Camada Independete de Dispositivo
int msqid_cti;				// ID da Fila de Mensagens com a Camada Tratamento de Interrupcoes
int shmid_regctrl;			// ID da Memoria Compartilhada dos Registradores da Controladora
int shmid_tabcfg;			// ID da Memoria Compartilhada da Tabela de Configuracao
int shmid_mpart;			// ID da Memoria Compartilhada da Matriz de Particoes
int semid_prod;				// ID do Semaforo de Exclusao Mutua entre Produtores
int semid_lista;			// ID do Semaforo de Exclusao Mutua da Lista de Requisicoes
int semid_contlista;			// ID do Semaforo Contador de Requisicoes da Lista de Requisicoes

int *VClustFinais;

float clustportril;		// clusters por trilha
float clustporcil;		// clusters por cilindro

typedef struct requisicao
	{				// ESTRUTURA DA REQUISICAO
		long int cluster;	// Indice do Cluster
		int operacao;		// Tipo da operacao na controladora, 0 (read), 1 (write), 2 (seek), 3 (recalibrate) e 4 (reset)
		struct timeval deadline;// Deadline da requisicao
		long int endmem;	// Endereco de memoria de destino do read ou origem do write
		int tamanho;		// Tamanho em bytes que serao manipulados na requisicao de read/write
		int pid;		// PID do processo que requisitou a operacao
	} requisicao_t;

typedef struct reqchs
{					// ESTRUTURA DE ELEMENTO DA LISTA
	requisicao_t req;		// Requisicao
	int C;			// Cilindro
	int H;			// Cabeca
	int S;			// Setor
	struct reqchs *next;	// Ponteiro para proximo elemento
} reqchs_t;

typedef struct tabcfg
	{				// ESTRUTURA DA TABELA DE PARTICAO
		int tmax_atend;		// Tempo maximo, em ms, que a controladora leva para executar uma operacao
		int numpart;		// Numero de particoes
		int numcabecas;		// Numero de cabecas
		int setporclu;		// Numero de Setores por Cluster
	} tabcfg_t;
tabcfg_t *tabcfg_addr;

typedef struct mpart
	{				// ESTUTURA DE UMA LINHA A MATRIZ DE PARTICOES
		int setportril;		// Numero de setores por trilha da particao
		int cilfinal;		// Cilindro Final da particao
	} mpart_t;
mpart_t *mpart_addr;

typedef struct regctrl
	{				// ESTRUTURA DOS REGISTRADORES DA CONTROLADORA
		int C;			// Numero do Cilindro
		int H;			// Numero da Cabeca
		int S;			// Numero do Setor
		int operacao;		// Tipo da operacao na controladora, 0 (read), 1 (write), 2 (seek), 3 (recalibrate) e 4 (reset)
		long int endmem;	// Endereco de memoria de destino do read ou origem do write
		int tamanho;		// Tamanho em bytes que serao manipulados na requisicao de read/write
		int Status;		// Status da opera�o
	} regctrl_t;
regctrl_t *regctrl_addr;

typedef struct msgbuf_req {			// BUFFER DA MENSAGEM DE REQUISICAO
	long mtype;				// Tipo da mensagem
	char mtext[sizeof(requisicao_t)];	// Conteudo da mensagem (requisicao)
} msgbuf_req_t;

typedef struct msgbuf_resp {			// BUFFER DA MENSAGEM DE RESPOSTA A REQUISICAO
	long mtype;				// Tipo da mensagem
	char mtext[1];				// Conteudo da mensagem (status)
} msgbuf_resp_t;

typedef struct msgbuf_int {		// BUFFER DA MENSAGEM DA CTI
	long mtype;				// Tipo da mensagem
} msgbuf_int_t;

reqchs_t * primeiro_lista; // Ponteiro para o primeiro elemento da lista de requisies

/*
 * Prototipo das Funcoes
 */

void geradorprod();
void produtor(reqchs_t *nova_requisicao);
void consumidor();
int modulo11(reqchs_t * nova_requisicao);
void modulo12(reqchs_t * nova_requisicao);
int modulo13(reqchs_t *reqchs);
int modulo14(reqchs_t *reqchs, int *contador_retry);
void monta_mensagem(int PID, int Status);
void calcular_chs(reqchs_t * requisicao);
int verificar_deadline(requisicao_t * requisicao);

/***********************************************************************************************************************************
	FUNCAO MAIN
***********************************************************************************************************************************/
int main(int argc, char *argv[])
{
	/*
	 * Declaracao das variaveis
	 */
	pthread_t gprod,cons;		// identificadores das threads
	int i;				// contador para for
	int cilinicial;			// cilindro inicial (para calculo do vetor de clusters finais)
	int contclusters;		// contador de clusters (para calculo do vetor de clusters finais)


	/*
	 * Declaracao das Filas de Mensagens
	 */
	msqid_cid = msgget(MSQ_CID,0666);
	msqid_cti = msgget(MSQ_CTI,0666);


	/*
	 * Declaracao das Memorias Compartilhadas dos Registradores da Controladora e da Tabela de Configuracao
	 */
	shmid_regctrl = shmget(SHM_REGCTRL,sizeof(regctrl_t),0666);
	shmid_tabcfg = shmget(SHM_TABCFG,sizeof(tabcfg_t),0666);


	/*
	 * Inicializando as Memorias Compartilhadas dos Registradores da Controladora e da Tabela de Configuracao
	 */
	regctrl_addr = (regctrl_t *) shmat(shmid_regctrl, NULL, 0);
	tabcfg_addr = (tabcfg_t *) shmat(shmid_tabcfg, NULL, 0);


	/*
	 * Declaracao da Memoria Compartilhada da Matriz de Particoes
	 */
	shmid_mpart = shmget(SHM_MPART,tabcfg_addr->numpart*sizeof(mpart_t),0666|IPC_CREAT);


	/*
	 * Inicializando a Memoria Compartilhada da Matriz de Particoes
	 */
	mpart_addr = (mpart_t *) shmat(shmid_mpart, NULL, 0);

	/*
	 * Cria e Inicia o Vetor de Cluster Finais
	 */
	VClustFinais = (int *) malloc(sizeof(int [tabcfg_addr->numpart]));	// Aloca memoria para o vetor de clusters finais
	cilinicial = 0;								// Seta cilindro inicial da primeira particao como 0
	contclusters = 0;							// Seta contador de clusters como 0
	for (i=0; i<tabcfg_addr->numpart; i++)					// Loop para cada particao
	{
		clustportril = (float)mpart_addr[i].setportril/(float)tabcfg_addr->setporclu;	// calcula clusters por trilha da particao i
		clustporcil = clustportril * (float) tabcfg_addr->numcabecas;			// calcula cluster por cilindro da particao i
		contclusters += (clustporcil * (mpart_addr[i].cilfinal-cilinicial+1))-1;	// aculuma o numero de clusters da particao no contador de clusters
		VClustFinais[i] = contclusters;							// atribui no vetor de clusters finais, o cluster final da particao i
		cilinicial = mpart_addr[i].cilfinal+1;						// atualiza o valor do cilindro inicial da proxima particao (caso for a ultima particao, o valor sera desconsiderado)
	}


	/*
	 * Declaracao dos Semaforos
	 */
	semid_prod = semget(SEM_PROD,1,0666|IPC_CREAT);
	semid_lista = semget(SEM_LISTA,1,0666|IPC_CREAT);
	semid_contlista = semget(SEM_CONTLISTA,1,0666|IPC_CREAT);

	/*
	 * Inicializando os semaforos de exlusao mutua
	 */
	semop(semid_prod,up,1);
	semop(semid_lista,up,1);

	/*
	 * Inicializando a Lista de Requisicoes
	 */
	primeiro_lista=NULL;


	/*
	 * Chamada das Threads
	 */
	pthread_create(&gprod,NULL,(void *)geradorprod,NULL);
	pthread_create(&cons,NULL,(void *)consumidor,NULL);

	/*
	 * Join do main nas Threads (Main espera as threads acabarem de executar)
	 */
	pthread_join(gprod,NULL);
}

/***********************************************************************************************************************************
	FUNCAO GERADORPROD
***********************************************************************************************************************************/
void geradorprod()
{
	// Variaveis da Funcao
	pthread_t prod;			// identificador da thread
	msgbuf_req_t msg_requisicao;	// buffer da mensagem com a requisicao
	requisicao_t * reqmsg_ptr;

	// Loop para receber requisicoes e chamar produtor
	while (1)
	{
		// Le mensagem de Requisicao da Camada Independente de Dispositivo (do tipo 1) com a requisicao
		msgrcv(msqid_cid,(msgbuf_req_t *)&msg_requisicao,sizeof(requisicao_t),1,0);
		reqmsg_ptr = (requisicao_t *) (msg_requisicao.mtext);
		reqchs_t * nova_requisicao = (reqchs_t*) malloc (sizeof (reqchs_t)); // aloca memoria para a nova requisicao
		nova_requisicao->req = *reqmsg_ptr; // copia a requisicao recebida para o endereco de memoria alocado

		// Chama Thread Produtor passando como argumento o ponteiro para a area de memoria alocada para a nova requisicao
		pthread_create(&prod,NULL,(void *)produtor,nova_requisicao);

	}
}

/***********************************************************************************************************************************
	FUNCAO PRODUTOR
***********************************************************************************************************************************/
void produtor ( reqchs_t * nova_requisicao)
{

 	// Variaveis da funcao
	msgbuf_resp_t msg_respostareq;	// buffer da mensagem de resposta a CID

	// Executa Analisar Requisicao e Calcular CHS
	//Se o deadline estiver estourado, entra no IF
	if(modulo11(nova_requisicao))
	{

		msg_respostareq.mtype = (nova_requisicao->req).pid; //seta o tipo da mensagem de resposta com o PID do processo solicitador da requisicao
		msg_respostareq.mtext[0] = (char) 4; // seta o campo de conteudo como 4, que identifica uma resposta caracterizada por deadline invalido
		msgsnd(msqid_cid,(msgbuf_resp_t *)&msg_respostareq,1,0);
		free(nova_requisicao);//libera a memoria alocada para a nova requisicao
		pthread_exit(NULL);

	}

	// Executa Armazenar e Ordenar Requisicao
	modulo12(nova_requisicao);

	#ifdef CASO_INT1112_LISTAVAZIAOBSERVAR
	// Cria arquivo texto
	FILE *saida;
	saida=fopen("caso_int1112_listavazia.txt","w");

 	reqchs_t * req_lista = primeiro_lista;
	int i=0;

	// Down no semaforo de exclusao mutua entre produtores
	semop(semid_prod, down, 1);
	// Down no semaforo de exclusao mutua entre produtor e consumidor
	semop(semid_lista, down, 1);
	while(req_lista != NULL)
	{
		fprintf(saida,"Requisicao %d da lista:\n\n-> C : %d\n-> H : %d\n-> S : %d\n-> Cluster: %d\n-> Operacao: %d\n-> Endereco de Memoria: %d\n-> Tamanho: %d\n-> PID: %d\n-> Deadline: %ds, %dus\n",i,req_lista->C,req_lista->H,req_lista->S,req_lista->req.cluster,req_lista->req.operacao,req_lista->req.endmem,req_lista->req.tamanho,req_lista->req.pid,req_lista->req.deadline.tv_sec,req_lista->req.deadline.tv_usec);
		fprintf(saida,"\n\n");
		i++;
		req_lista=req_lista->next;
	}
	// Up no semaforo de exclusao mutua entre produtor e consumidor
	semop(semid_lista, up, 1);
	// Up no semaforo de exclusao mutua entre produtores
	semop(semid_prod, up, 1);
	fclose(saida);
	#endif

	#ifdef CASO_INT1112_INICIOLISTAOBSERVAR
	// Cria arquivo texto
	FILE *saida;
	saida=fopen("caso_int1112_iniciolista.txt","w");

 	reqchs_t * req_lista = primeiro_lista;
	int i=0;

	// Down no semaforo de exclusao mutua entre produtores
	semop(semid_prod, down, 1);
	// Down no semaforo de exclusao mutua entre produtor e consumidor
	semop(semid_lista, down, 1);
	while(req_lista != NULL)
	{
		fprintf(saida,"Requisicao %d da lista:\n\n-> C : %d\n-> H : %d\n-> S : %d\n-> Cluster: %d\n-> Operacao: %d\n-> Endereco de Memoria: %d\n-> Tamanho: %d\n-> PID: %d\n-> Deadline: %ds, %dus\n",i,req_lista->C,req_lista->H,req_lista->S,req_lista->req.cluster,req_lista->req.operacao,req_lista->req.endmem,req_lista->req.tamanho,req_lista->req.pid,req_lista->req.deadline.tv_sec,req_lista->req.deadline.tv_usec);
		fprintf(saida,"\n\n");
		i++;
		req_lista=req_lista->next;
	}
	// Up no semaforo de exclusao mutua entre produtor e consumidor
	semop(semid_lista, up, 1);
	// Up no semaforo de exclusao mutua entre produtores
	semop(semid_prod, up, 1);
	fclose(saida);
	#endif

	#ifdef CASO_INT1112_FIMLISTAOBSERVAR
	// Cria arquivo texto
	FILE *saida;
	saida=fopen("caso_int1112_fimlista.txt","w");

 	reqchs_t * req_lista = primeiro_lista;
	int i=0;

	// Down no semaforo de exclusao mutua entre produtores
	semop(semid_prod, down, 1);
	// Down no semaforo de exclusao mutua entre produtor e consumidor
	semop(semid_lista, down, 1);
	while(req_lista != NULL)
	{
		fprintf(saida,"Requisicao %d da lista:\n\n-> C : %d\n-> H : %d\n-> S : %d\n-> Cluster: %d\n-> Operacao: %d\n-> Endereco de Memoria: %d\n-> Tamanho: %d\n-> PID: %d\n-> Deadline: %ds, %dus\n",i,req_lista->C,req_lista->H,req_lista->S,req_lista->req.cluster,req_lista->req.operacao,req_lista->req.endmem,req_lista->req.tamanho,req_lista->req.pid,req_lista->req.deadline.tv_sec,req_lista->req.deadline.tv_usec);
		fprintf(saida,"\n\n");
		i++;
		req_lista=req_lista->next;
	}
	// Up no semaforo de exclusao mutua entre produtor e consumidor
	semop(semid_lista, up, 1);
	// Up no semaforo de exclusao mutua entre produtores
	semop(semid_prod, up, 1);
	fclose(saida);
	#endif


	#ifdef CASO_INT1112_MEIOLISTAOBSERVAR
	// Cria arquivo texto
	FILE *saida;
	saida=fopen("caso_int1112_meiolista.txt","w");

 	reqchs_t * req_lista = primeiro_lista;
	int i=0;

	// Down no semaforo de exclusao mutua entre produtores
	semop(semid_prod, down, 1);
	// Down no semaforo de exclusao mutua entre produtor e consumidor
	semop(semid_lista, down, 1);
	while(req_lista != NULL)
	{
		fprintf(saida,"Requisicao %d da lista:\n\n-> C : %d\n-> H : %d\n-> S : %d\n-> Cluster: %d\n-> Operacao: %d\n-> Endereco de Memoria: %d\n-> Tamanho: %d\n-> PID: %d\n-> Deadline: %ds, %dus\n",i,req_lista->C,req_lista->H,req_lista->S,req_lista->req.cluster,req_lista->req.operacao,req_lista->req.endmem,req_lista->req.tamanho,req_lista->req.pid,req_lista->req.deadline.tv_sec,req_lista->req.deadline.tv_usec);
		fprintf(saida,"\n\n");
		i++;
		req_lista=req_lista->next;
	}
	// Up no semaforo de exclusao mutua entre produtor e consumidor
	semop(semid_lista, up, 1);
	// Up no semaforo de exclusao mutua entre produtores
	semop(semid_prod, up, 1);
	fclose(saida);
	#endif
}

/***********************************************************************************************************************************
	FUNCAO CONSUMIDOR
***********************************************************************************************************************************/
void consumidor()
{
	// Variaveis da funcao
	msgbuf_int_t msg_cti;		// buffer da mensagem da CTI
	int retry;			// verificador de retry
	int contador_retry;		// contador de retry
	reqchs_t reqchs;		// estrutura de um elemento da lista de requisicoes
	msgbuf_resp_t msg_respostareq;	// buffer da mensagem de resposta a CID

	// Loop para atender requisicao e verificar status e tratar erros
	while (1)
	{
		// Executa Atender Requisicao
		if(modulo13(&reqchs))
		{

			msg_respostareq.mtype = reqchs.req.pid;
   			msg_respostareq.mtext[0] = (char) 4;
			msgsnd(msqid_cid,(msgbuf_resp_t *)&msg_respostareq,1,0);

			free(reqchs);

			continue;
		}

		//Contador para o caso do retry
		contador_retry=0;

		// Loop de repeticoes no caso de retries
		retry=0;
		do{
			// Recebe mensagem da rotina da Camada de Tratamento de Interrupcoes (do tipo 1) quando a Controladora acabar de executar a operacao
			msgrcv(msqid_cti,(msgbuf_int_t *)&msg_cti,0,1,0);

			// Executa Verificar Status e Tratar Erros
			retry = modulo14( &reqchs , &contador_retry);

		} while (retry == 1);
	}
}



/***********************************************************************************************************************************
	FUNCAO MODULO11
***********************************************************************************************************************************/
int modulo11(reqchs_t * nova_requisicao)
{

 	//Chamada da funcao verificar_deadline
	if(verificar_deadline(&nova_requisicao->req))
	{
		return 1;
	}

	//Chamada da funcao calcular CHS
	calcular_chs(nova_requisicao);

	return 0;
}


/***********************************************************************************************************************************
	FUNCAO MODULO12
***********************************************************************************************************************************/
void modulo12(reqchs_t * nova_requisicao)
{

	reqchs_t * req_auxiliar = NULL;//variavel auxiliar para percorrer a lista
	reqchs_t * req_auxiliar2 = NULL;//variavel auxiliar para percorrer a lista

	// Down no semaforo de exclusao mutua entre produtores
	semop(semid_prod, down, 1);

	// Down no semaforo de exclusao mutua entre produtor e consumidor
	semop(semid_lista, down, 1);


	//Variaveis auxiliares para percorrer a lista de requisicoes
	req_auxiliar=primeiro_lista;
	req_auxiliar2=NULL;

	/*Se a lista estiver vazia, insere o elemento e aponta o indicador de primeiro elemento da lista para ele*/
	if(req_auxiliar == NULL)
	{
		req_auxiliar = nova_requisicao;
		primeiro_lista = req_auxiliar;

	}
	else
	{

	   /*Percorre a lista ateh o final*/
	   while(req_auxiliar != NULL)
	   {
	      /*Se o deadline da requisicao a ser inserida for menor que o da requisicao da lista atual (comparando o campo de segundos), insere o elemento na posicao do elemento atual da lista*/
  	      if ((nova_requisicao->req).deadline.tv_sec < (req_auxiliar -> req).deadline.tv_sec)
	      {
	      	  /*Se o elemento for inserido no inicio da lista, atualiza o indicar de primeiro elemento da lista*/
	      	  if(req_auxiliar==primeiro_lista)
		  {
		     nova_requisicao->next=primeiro_lista;
		     primeiro_lista = nova_requisicao;
		     break;
		  }
		  else
		  {
		     req_auxiliar2->next=nova_requisicao;
		     nova_requisicao->next=req_auxiliar;
		     break;
		  }
	      }
	      else
	      /*Se o deadline da requisicao a ser inserida for menor que o da requisicao da lista atual (comparando o campo de microsegundos), insere o elemento na posicao do elemento atual da lista*/
		if((nova_requisicao->req).deadline.tv_sec == (req_auxiliar -> req).deadline.tv_sec &&
		  (nova_requisicao->req).deadline.tv_usec < (req_auxiliar -> req).deadline.tv_usec)
		{
		     /*Se o elemento for inserido no inicio da lista, atualiza o indicar de primeiro elemento da lista*/
		     if(req_auxiliar==primeiro_lista)
		     {
		        nova_requisicao->next=primeiro_lista;
		        primeiro_lista = nova_requisicao;
		        break;
		     }
		     else
		     {
		        req_auxiliar2->next=nova_requisicao;
		        nova_requisicao->next=req_auxiliar;
		        break;
		     }
		}

		//Anda com os ponteiros auxiliares
		req_auxiliar2=req_auxiliar;
		req_auxiliar=req_auxiliar->next;

	   }

	   /*Caso o deadline da requisicao a ser inserida seja maior que todos os elementos da lista, este serah inserido no final da mesma e seu ponteiro apontado para nulo*/
	   if(req_auxiliar==NULL)
	   {
	      req_auxiliar2->next=nova_requisicao;
	      nova_requisicao->next=NULL;
	   }
	}


	// Up no semaforo contador de requisicoes da lista de requisicoes
	semop(semid_contlista, up, 1);

	// Up no semaforo de exclusao mutua entre produtor e consumidor
	semop(semid_lista, up, 1);

	// Up no semaforo de exclusao mutua entre produtores
	semop(semid_prod, up, 1);

}


/***********************************************************************************************************************************
	FUNCAO MODULO13
***********************************************************************************************************************************/
int modulo13(reqchs_t *reqchs)
{
	//Variaveis da Funcao
	struct timeval deadlinemaximo;	// deadline maximo em que a requisicao deve ser atendida


	// Down no semaforo contador de requisicoes da lista de requisicoes
	semop(semid_contlista, down, 1);


	// Down no semaforo de exclusao mutua da lista
	semop(semid_lista, down, 1);

	// Retira requisicao da lista de requisicoes
	*reqchs = *primeiro_lista;
	free(primeiro_lista);
	primeiro_lista = reqchs->next;

	// Up no semaforo de exclusao mutua da lista
	semop(semid_lista, up, 1);

	#ifdef CASO_INT111213_REMOCAOLISTAOBSERVAR
	// Cria arquivo texto
	FILE *saida;
	saida=fopen("caso_int111213_remocaolista.txt","w");

	fprintf(saida,"Requisicao removida da lista:\n\n-> C : %d\n-> H : %d\n-> S : %d\n-> Cluster: %d\n-> Operacao: %d\n-> Endereco de Memoria: %d\n-> Tamanho: %d\n-> PID: %d\n-> Deadline: %ds, %dus\n",reqchs->C,reqchs->H,reqchs->S,reqchs->req.cluster,reqchs->req.operacao,reqchs->req.endmem,reqchs->req.tamanho,reqchs->req.pid,reqchs->req.deadline.tv_sec,reqchs->req.deadline.tv_usec);
	fclose(saida);
	#endif

	#ifdef CASO_INT111213_DEADLINEESTOURADOCAUSAR
	sleep(5);
	#endif

	// Verifica Deadline
	gettimeofday( &deadlinemaximo, NULL );	// obtem hora atual; deadline maximo eh igual a hora atual...

	deadlinemaximo.tv_usec += tabcfg_addr->tmax_atend*1000;	// ...mais o tempo tempo maximo de atendimento
	if (deadlinemaximo.tv_usec >= 1000000)	// verifica se usec ja ultrapassa um sec, se sim ajusta os campos
	{
		deadlinemaximo.tv_sec++;
		deadlinemaximo.tv_usec -= 1000000;
	}
	if (deadlinemaximo.tv_sec > reqchs->req.deadline.tv_sec || (deadlinemaximo.tv_sec == reqchs->req.deadline.tv_sec && deadlinemaximo.tv_usec > reqchs->req.deadline.tv_usec))
	{
		
		return 1;
	}

	#ifdef CASO_INT111213_DEADLINEVALIDOOBSERVAR
	// Cria arquivo texto
	FILE *saida;
	saida=fopen("caso_int111213_deadlinevalido.txt","w");

	fprintf(saida,"Requisicao com deadline valido:\n\n-> C : %d\n-> H : %d\n-> S : %d\n-> Cluster: %d\n-> Operacao: %d\n-> Endereco de Memoria: %d\n-> Tamanho: %d\n-> PID: %d\n-> Deadline: %ds, %dus\n",reqchs->C,reqchs->H,reqchs->S,reqchs->req.cluster,reqchs->req.operacao,reqchs->req.endmem,reqchs->req.tamanho,reqchs->req.pid,reqchs->req.deadline.tv_sec,reqchs->req.deadline.tv_usec);
	fclose(saida);
	#endif

	// Ativa Controladora
	regctrl_addr->C = reqchs->C;
	regctrl_addr->H = reqchs->H;
	regctrl_addr->S = reqchs->S;
	regctrl_addr->endmem = reqchs->req.endmem;
	regctrl_addr->tamanho = reqchs->req.tamanho;
	regctrl_addr->operacao = reqchs->req.operacao;

	return 0;
}


/***********************************************************************************************************************************
	FUNCAO MODULO14
***********************************************************************************************************************************/
int modulo14(reqchs_t *reqchs, int *contador_retry)
{
	/*
	 * Verificar Status 
	 */
	switch(regctrl_addr->Status)
	{
		case 1 :
		 {					// Erro transiente de Digito de Controle
		 	if (*contador_retry == 2)	// Nota-se o erro permanente de digito de controle
			{
				#ifdef TESTES
				printf("case 1, Erro permanente\n");
				#endif
				regctrl_addr->Status=2;	// Status 2 indica erro permanente de digito de controle
				monta_mensagem(reqchs->req.pid, regctrl_addr->Status); //chama rotina que monta e manda mensagem.
				return 0;	// retorna valor que sai do loop de retry
			}
			else				// Tentativa de retry
			{
				reqchs_t *proxima_req = (reqchs_t *) malloc(sizeof (reqchs_t));	// Declara ponteiro para proxima requisicao
				proxima_req = primeiro_lista;	// Pega proximo elemento a ser atendido
				if((proxima_req!=NULL) && ((proxima_req->req.deadline.tv_sec < reqchs->req.deadline.tv_sec)||((proxima_req->req.deadline.tv_sec == reqchs->req.deadline.tv_sec) &&(proxima_req->req.deadline.tv_usec < reqchs->req.deadline.tv_usec))))	// compara requisicao atual com proxima requisicao
				{			// proxima requisicao e mais prioritaria que atual
					#ifdef TESTES
					printf("case 1, Req + prioritaria\n");
					#endif
					monta_mensagem(reqchs->req.pid, regctrl_addr->Status); //chama rotina que monta e manda mensagem.
					free(proxima_req);	// libera memoria alocada para a comparao
					*contador_retry = 0;	// zera o contador de retry
					return 0;	// retorna valor que sai do loop de retry
				}
				else
				{
					// proxima requisicao e menos prioritaria que a atual
					free(proxima_req);
					if(!verificar_deadline(&reqchs->req))		// verifica deadline
					{
						//deadline valido
						*contador_retry=*contador_retry+1;			// incrementa contador de retries
						#ifdef TESTES
						printf("contador retry= %d\n",*contador_retry);
						#endif
						regctrl_addr->operacao = reqchs->req.operacao;	// reativar controladora
						return 1; 						// retorna valor que continua o loop de retry
					}
					#ifdef TESTES
					printf("case 1, Deadline Estourado\n");
					#endif
					// deadline estourado
					monta_mensagem(reqchs->req.pid, 4);	//chama rotina que monta e manda mensagem.
					return 0;							// retorna valor que sai do loop de retry
				}
			}
		}
		case 3 : 
		{
			if (*contador_retry == 2)	// Nota-se o erro de seek nao se
			{
				#ifdef TESTES
				printf("case 3, Erro de seek 3x\n");
				#endif
				monta_mensagem(reqchs->req.pid, regctrl_addr->Status); //chama rotina que monta e manda mensagem.
				return 0;	// retorna valor que sai do loop de retry
			}
			// Erro de seek
			msgbuf_int_t msg_cti;		// Mensagem para receber mensagem da CTI
			regctrl_addr->operacao=3;	// Ativar a controladora com 3, recalibrate
			msgrcv(msqid_cti,(msgbuf_int_t *)&msg_cti,0,1,0);	// Recebe mensagem de fim de recalibrate cd CTI
			
			//Verificar requisições existentes para tentar retry
			reqchs_t *proxima_req = (reqchs_t *) malloc(sizeof (reqchs_t));	// Declara ponteiro para proxima requisicao
			proxima_req = primeiro_lista;	// Pega proximo elemento a ser atendido
			
			// compara requisicao atual com proxima requisicao
			if((proxima_req!=NULL) && ((proxima_req->req.deadline.tv_sec < reqchs->req.deadline.tv_sec)||((proxima_req->req.deadline.tv_sec == reqchs->req.deadline.tv_sec) &&(proxima_req->req.deadline.tv_usec < reqchs->req.deadline.tv_usec))))	// compara requisicao atual com proxima requisicao
			{
				#ifdef TESTES
				printf("case 3, Req + prioritaria\n");
				#endif
				// proxima requisicao e mais prioritaria que atual
				monta_mensagem(reqchs->req.pid, regctrl_addr->Status);	//chama rotina que monta e manda mensagem.
				free(proxima_req);	// libera memoria alocada para requisio
			}
			else
			{
				// proxima requisicao e menos prioritaria que a atual
				free(proxima_req);
				if(!verificar_deadline(&reqchs->req))			// verifica deadline
				{
					*contador_retry=*contador_retry+1;			// incrementa contador de retries
					#ifdef TESTES
					printf("contador retry= %d\n",*contador_retry);
					#endif
					regctrl_addr->operacao = reqchs->req.operacao;	// reativar controladora
					return 1;	// retorna valor que continua o loop de retry
				}
				#ifdef TESTES
				printf("case 3, Deadline Estourado\n");
				#endif
				monta_mensagem(reqchs->req.pid, 4);
				return 0; 	// retorna valor que sai do loop de retry
			} 
		}
		default:				// para operacao ok 
		{					// Operacao concliuda com exito
			#ifdef TESTES
			printf("Default\n");
			#endif
			*contador_retry = 0;	//zera contador de retries
			monta_mensagem( (reqchs->req).pid , regctrl_addr->Status );	//chama rotina que monta e manda mensagem. reqchs->req.pid
			return 0;			// retorna valor que sai do loop de retry
		}
	}
}
void monta_mensagem(int PID, int Status)
{
	int retorno_msgsnd=-1;
	msgbuf_resp_t mensagem_resposta;		// Declara mensagem de retorno
	mensagem_resposta.mtype=(long)PID;		// Atribui valor do PID ao tipo da mensagem
	mensagem_resposta.mtext[0]=Status;	// Atribui valor do Status a mensagem
	retorno_msgsnd = msgsnd(msqid_cid, (msgbuf_resp_t *)&mensagem_resposta, 1, 0);	// Envia mensagem a CID.
	//printf("Foi enviafo Status= %d\n",Status);
	return;
}

int verificar_deadline(requisicao_t * requisicao)
{


	struct timeval deadlinemaximo;//variavel com a soma da hora atual e o tempo maximo de atendimento contido na tabela de configuracao


	// Obtem a hora atual
	gettimeofday( &deadlinemaximo, NULL );
	deadlinemaximo.tv_usec += (tabcfg_addr->tmax_atend * 1000);//adiciona a hora atual o tempo maximo de atendimento

	if (deadlinemaximo.tv_usec >= 1000000)// verifica se usec ja ultrapassa um sec, se sim ajusta os campos
	{
		deadlinemaximo.tv_sec++;
		deadlinemaximo.tv_usec -= 1000000;
	}

	//Verifica se a requisicao tem deadline menor que o do deadlinemaximo;caso positivo, retorna 1, estouro de deadline
	//Caso contrario, retorna 0, deadline da requisicao eh valido
	if(deadlinemaximo.tv_sec > requisicao->deadline.tv_sec)
		return 1;
	else
	{
		if(deadlinemaximo.tv_sec == requisicao->deadline.tv_sec &&
		   deadlinemaximo.tv_usec > requisicao->deadline.tv_usec)
		   return 1;
		else
		   return 0;
	}
}

void calcular_chs(reqchs_t * requisicao)
{

	long int indice_cluster = (requisicao->req).cluster;//indice do cluster da requisicao
	int particao=0;//particao onde se encontra o cluster
	float clusterportril=0;//quantidade de clusters por trilha
	int cluster_particao=0;//indice do cluster referente a particao;em cada particao, os cluster sao numerados a partir do 0
	float clusterporcil = 0;//quantidade de clusters por cilindro
	int cilindro_particao=0;//cilindro referente a particao;em cada particao os cilindros sao numerados a partir do 0
	int c = 0;// cilindro calculado
	int h = 0;//cabeca calculada
	int s = 0;//setor calculado

 	float resto=0;

 	//Obtem a particao em que se encontra o cluster atraves do Vetor de Clusters Finais de cada Particao
	for(particao=0;particao<tabcfg_addr->numpart;)
	{
		if(indice_cluster<VClustFinais[particao] || indice_cluster==VClustFinais[particao])
			break;
		particao++;
	}


	//Calcula cluster por trilha e por cilindro para utilizacao nos calculos futuros
	clusterportril = (float)mpart_addr[particao].setportril / (float)tabcfg_addr->setporclu;
	clusterporcil = clusterportril * (float)tabcfg_addr->numcabecas;


	/*Caso a particao seja a primeira, o indice do cluster referente a particao eh o proprio
	  indice do cluster passado na requisicao.
	  Caso contrario, o indice do cluster referente a particao eh indice do cluster passado
	  na requisicao menos o indice do cluster final da particao anterior menos 1.
	*/
	if(particao!=0)
		cluster_particao = indice_cluster - VClustFinais[particao-1] -1;
	else
		cluster_particao = indice_cluster;

	//Obtencao do cilindro referente a particao
	cilindro_particao = (int) ((float)cluster_particao / clusterporcil);

	/*Caso a particao seja a primeira, o cilindro referente a particao eh o cilindro calculado.
	  Caso contrario, o cilindro referente a particao eh cilindro calculado
	  mais o cilindro final da particao anterior somado com 1.
	*/
	if(particao!=0)
		c = cilindro_particao + mpart_addr[particao-1].cilfinal + 1;
	else
		c = cilindro_particao;

	//Obtencao do resto da divisao entre cluster_particao e clusterporcil
	resto = (float)cluster_particao - ((float)cilindro_particao * clusterporcil);

	//Obtencao da cabeca
	h = (int)(resto / clusterportril);

	//Obtencao do resto da divisao entre (cluster_particao dividido por clusterporcil) e clusterportril
	resto = ((float)resto - ((float)h * clusterportril));

	//Obtecao dos setor
	s = resto * tabcfg_addr->setporclu;

	//Atribuicao do calculo para as variaveis C,H,S da requisicao
	requisicao->C = c;
	requisicao->H = h;
	requisicao->S = s;

}

